#!/usr/bin/env python
"""
Tests for yamlstratus
"""
import json
import os
import os.path
import yamlstratus

test_dir = 'test'


def compare_json(expected, actual):
    """
    Compares to dictionaries representing json objects
    :param expected: The expected json
    :param actual: The actual json
    :return: where the two json representations are equal
    """
    if isinstance(expected, dict):
        if not isinstance(actual, dict):
            assert actual == expected
        assert isinstance(actual, dict)
        assert len(expected) == len(actual)
        for expected_key in expected:
            assert expected_key in actual
            compare_json(expected[expected_key], actual[expected_key])
    elif isinstance(expected, list):
        assert isinstance(actual, list)
        assert len(expected) == len(actual)
        cnt = 0
        while cnt < len(expected):
            compare_json(expected[cnt], actual[cnt])
            cnt += 1
    else:
        assert expected == actual


def get_test_pairs(dir):
    """
    Pulls out pairs of files corresponding to a test yaml file and ab
    expected json file
    :param dir: the directory in which to search for such pairs
    :return: iterator over test pairs
    """
    for filename in os.listdir(dir):
        if not filename.endswith('.json'):
            continue
        # .yaml named same as corresponding .json
        yield (os.path.join(test_dir, filename),
               os.path.join(test_dir, filename[:-5] + '.yaml'))


def test_load_all():
    """Tests yamlstratus.load_all()"""
    for json_filename, yaml_filename in get_test_pairs(test_dir):
        with open(json_filename, 'r') as json_f:
            all_expected_json = json.loads(json_f.read())
        with open(yaml_filename, 'r') as yaml_f:
            json_cnt = 1
            for actual_json in yamlstratus.load_all(
                    yaml_f,
                    **all_expected_json[0]):
                expected_json = all_expected_json[json_cnt]
                compare_json(expected_json, actual_json)
                json_cnt += 1


def test_load_all_as_json():
    """Tests yamlstratus.load_all_as_json()"""
    for json_filename, yaml_filename in get_test_pairs(test_dir):
        with open(json_filename, 'r') as json_f:
            all_expected_json = json.loads(json_f.read())
        with open(yaml_filename, 'r') as yaml_f:
            json_cnt = 1
            for actual_json_str in yamlstratus.load_all_as_json(
                    yaml_f,
                    **all_expected_json[0]):
                actual_json = json.loads(actual_json_str)
                expected_json = all_expected_json[json_cnt]
                compare_json(expected_json, actual_json)
                json_cnt += 1


def test_to_json():
    """Tests yamlstratus.load_all()"""
    for json_filename, yaml_filename in get_test_pairs(test_dir):
        with open(json_filename, 'r') as json_f:
            all_expected_json = json.loads(json_f.read())
        with open(yaml_filename, 'r') as yaml_f:
            json_cnt = 1
            for actual_json in yamlstratus.load_all(
                    yaml_f,
                    **all_expected_json[0]):
                actual_json_str = yamlstratus.to_json(actual_json)
                actual_json = json.loads(actual_json_str)
                expected_json = all_expected_json[json_cnt]
                compare_json(expected_json, actual_json)
                json_cnt += 1


def test_load():
    """Tests yamlstratus.load()"""
    for json_filename, yaml_filename in get_test_pairs(test_dir):
        with open(json_filename, 'r') as json_f:
            all_expected_json = json.loads(json_f.read())
        with open(yaml_filename, 'r') as yaml_f:
            json_cnt = 1
            # Split yaml file into parts, testing each individually
            for yaml in yaml_f.read().split('---')[1:]:
                actual_json = yamlstratus.load(yaml, **all_expected_json[0])
                expected_json = all_expected_json[json_cnt]
                compare_json(expected_json, actual_json)
                json_cnt += 1


def test_load_as_json():
    """Tests yamlstratus.load_as_json()"""
    for json_filename, yaml_filename in get_test_pairs(test_dir):
        with open(json_filename, 'r') as json_f:
            all_expected_json = json.loads(json_f.read())
        with open(yaml_filename, 'r') as yaml_f:
            json_cnt = 1
            # Split yaml file into parts, testing each individually
            for yaml in yaml_f.read().split('---')[1:]:
                actual_json_str = yamlstratus.load_as_json(
                    yaml,
                    **all_expected_json[0])
                actual_json = json.loads(actual_json_str)
                expected_json = all_expected_json[json_cnt]
                compare_json(expected_json, actual_json)
                json_cnt += 1



"""
Provides functionality for converting YAML formatted files to JSON.

Specifically, YAML incorporating the YAML Stratus format is handled.

Function Exports:

- load()
- to_json()
- load_as_json()
- load_all()
- load_all_as_json()
"""

import json

from .yamlstratus import YamlStratus

__version__ = '1.0'


def load(stream, root_tag='main', include_dirs=None, params=None):
    """
    Parse a YAML stratus stream returning a dictionary

    :param stream: stream over the YAML Stratus document
    :param root_tag: the name of the root node in the document to return
        (default 'main')
    :param include_dirs: list of search paths for includes
    :param params: dictionary of parameter name-values
    :rtype: dictionary
    """
    objs = YamlStratus(include_dirs=include_dirs, params=params).load(stream)
    if root_tag is None:
        return objs
    elif objs is None:
        return None
    elif isinstance(objs, str) and objs == root_tag:
        return {}
    elif not isinstance(objs, dict):
        raise Exception('Root node is not a dictionary')
    elif root_tag in objs:
        return objs[root_tag]
    else:
        return None


def load_all(stream, root_tag='main', include_dirs=None, params=None):
    """
    Stream in multiple documents in YAML Stratus format returning an iterator
    that yields corresponding dictionaries

    :param stream: stream over the YAML Stratus document
    :param root_tag: the name of the root node in the document to return
        (default 'main')
    :param include_dirs: list of search paths for includes
    :param params: dictionary of parameter name-values
    :rtype: iterator over dictionaries
    """
    for objs in YamlStratus(include_dirs=include_dirs, params=params).load_all(
            stream):
        if root_tag is None:
            yield objs
        elif objs is None:
            yield None
        elif isinstance(objs, str) and objs == root_tag:
            yield {}
        elif not isinstance(objs, dict):
            raise Exception('Root node is not a dictionary {0}'.format(objs))
        elif root_tag in objs:
            yield objs[root_tag]
        else:
            raise Exception(
                'No node marked "{0}" in document root'.format(root_tag))


def to_json(objs):
    """
    Convenience method for returning am object as JSON

    :param objs: any python dictionary though meant for one generated
        by YamlStratus
    :rtype: string in JSON format
    """
    return json.dumps(objs, indent=4, separators=(',', ': '))


def load_as_json(stream, root_tag='main', include_dirs=None, params=None):
    """
    Stream in YAML Stratus stream and returning as JSON

    :param stream: stream over the YAML Stratus document
    :param root_tag: the name of the root node in the document to return
        (default 'main')
    :param include_dirs: list of search paths for includes
    :param params: dictionary of parameter name-values
    :rtype: string in JSON format
    """
    objs = load(stream, include_dirs=include_dirs, params=params,
                root_tag=root_tag)
    return to_json(objs) if not objs is None else []


def load_all_as_json(stream, include_dirs=None, params=None, root_tag='main'):
    """
    Stream in multiple documents in YAML Stratus format returning as iterator
    yielding corresponding JSON documents

    :param stream: stream over the YAML Stratus document
    :param root_tag: the name of the root node in the document to return
        (default 'main')
    :param include_dirs: list of search paths for includes
    :param params: dictionary of parameter name-values
    :rtype: iterator over JSON formatted strings
    """
    for objs in load_all(stream, include_dirs=include_dirs, params=params,
                         root_tag=root_tag):
        yield to_json(objs)



#!/usr/bin/env python
"""
Script or converting YAML formatted files to JSON. Specifically,
YAML incorporating the YAML Stratus format is handled.
"""
import os

import argparse

import yamlstratus

desc = "Convert YAML Stratus format to json"

parser = argparse.ArgumentParser(description=desc)

parser.add_argument("src",
                    metavar="src",
                    nargs='+',
                    help="the name of the source file")

parser.add_argument("-i",
                    "--include-dirs",
                    help="the search path of include files")

parser.add_argument("-r",
                    "--root-tag",
                    help="the source tag that forms the root of the " +
                         "returned document")

parser.add_argument("-o",
                    "--output-dir",
                    help="the destination directory of generated file")

parser.add_argument('-P',
                    "--param",
                    nargs=1,
                    action="append",
                    help="Parameter to set. Ex -P LoadBalancerDnsWeight=5")

args = parser.parse_args()

incdirs = (args.include_dirs or ".").split(os.pathsep)
outdir = args.output_dir or "target"
params = args.param or []
root_tag = args.root_tag or "main"

params_dict = dict()
for param in params:
    param_name, param_value = param[0].split('=')
    params_dict[param_name] = param_value

if not os.path.exists(outdir):
    os.makedirs(outdir)

for filename in args.src:
    output_file_name = os.path.splitext(os.path.basename(filename))[0] + '.json'
    if outdir is not None:
        output_file_name = os.path.join(outdir, output_file_name)

    with open(output_file_name, 'w') as out:
        with open(filename, 'r') as f:
            file_incdirs = incdirs + [os.path.dirname(filename)]
            for json_part in yamlstratus.load_all_as_json(
                    f,
                    include_dirs=file_incdirs,
                    params=params_dict,
                    root_tag=root_tag):
                out.write(json_part)



from __future__ import print_function
import os
import sys

from setuptools import setup
from setuptools.command.test import test as TestCommand

import yamlstratus

here = os.path.abspath(os.path.dirname(__file__))

long_description = """
YAML Stratus provides extensions to YAML that make it amenable to use with AWS
CloudFormation, as well as a python binding,
and a tool for converting to the standard JSON used by CloudFormation.

Standard YAML capabilities not part of JSON:

* comments
* back references
* block literals

Note also that JSON is YAML, so familiar JSON data can be embedded within YAML.

Extension to YAML provided by YAML Stratus:

* !include - YAML files can include other YAML files
* !param - YAML files can include parameters
* !merge - Allows the merging of data from two source data hierarchies
"""


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        errcode = pytest.main(self.test_args)
        sys.exit(errcode)


setup(
    name='yamlstratus',
    scripts=['scripts/ystratus.py'],
    version=yamlstratus.__version__,
    url='http://github.com/shannonwhite/yamlstratus/',
    download_url='https://github.com/kikinteractive/yaml-stratus/tarball/0.1',
    license='Apache Software License',
    author='Shannon White',
    tests_require=['pytest'],
    cmdclass={'test': PyTest},
    author_email='shannon.white@kik.com',
    description='Python for yamlstratus builder',
    long_description=long_description,
    packages=['yamlstratus'],
    include_package_data=True,
    platforms='any',
    test_suite='yamlstratus.test.test_yamlstratus',
    classifiers=[
        'Programming Language :: Python',
        'Natural Language :: English',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        "Topic :: Text Processing :: Markup",
    ],
    extras_require={
        'testing': ['pytest'],
    }
)



#!/usr/bin/env python

"""
Provides functionality for converting YAML formatted files to JSON.

Specifically, YAML incorporating the YAML Stratus format is handled.

Class Exports:

- YamlStratusLoader
- YamlStratus
"""

import base64
import os.path

import yaml


class MergeNode(object):
    """Represents a node in YAML document being merged"""

    def __init__(self, starting_from, merge_with):
        self.starting_from = starting_from
        self.merge_with = merge_with
        self.is_merged = False
        # The result of the merge should be typed like the starting node
        self.target = [] if isinstance(starting_from, list) or isinstance(
            starting_from, str) else dict()


class YamlStratusLoader(yaml.Loader):
    """
    Custom loader for handling custom extensions
    """

    def __init__(self, stream, params, include_dirs):
        super(YamlStratusLoader, self).__init__(stream)
        self.include_dirs = include_dirs or []
        self.merge_nodes = []
        self.removed_node = object()
        self.replaced_nodes = []
        self.params_dict = params or dict()

    @staticmethod
    def __check_include_file(directory, filename):
        """
        For finding include file in given directory. Tries various suffixes
        and no suffix
        """
        path = os.path.join(directory, filename) + '.yaml'
        if os.path.exists(path):
            return path
        path = os.path.join(directory, filename) + '.yml'
        if os.path.exists(path):
            return path
        path = os.path.join(directory, filename)
        if os.path.exists(path):
            return path
        return None

    def __find_include_file(self, filename):
        """Searches for file in the include path"""
        path = None
        if self.include_dirs is None or len(self.include_dirs) == 0:
            path = YamlStratusLoader.__check_include_file('.', filename)
        else:
            for include_dir in self.include_dirs:
                path = YamlStratusLoader.__check_include_file(include_dir,
                                                              filename)
                if path is not None:
                    break
        if path is None:
            raise Exception("Cannot find {0}".format(filename))
        return path

    def include(self, node):
        """
        Extension that handles including of other yaml files
        """
        name = self.construct_scalar(node)

        index = name.find('.')

        if index == -1:
            filename = name
            node_name = None
        else:
            filename = name[:index]
            node_name = name[index + 1:]
        path = self.__find_include_file(filename)
        with open(path, 'r') as include_fp:
            included_yaml_stratus = YamlStratus(include_dirs=self.include_dirs,
                                                params=self.params_dict)
            obj = included_yaml_stratus.load(include_fp)
            if node_name is None:
                return obj
            else:
                for node in node_name.split('.'):
                    obj = obj[node]
                return obj

    def include_base64(self, node):
        """
        Extension that handles including of other file as a base64 encoded
        string
        """
        filename = self.construct_scalar(node)

        path = self.__find_include_file(filename)
        with open(path, 'r') as include_fp:
            return base64.b64encode(include_fp.read())

    def remove(self, node):
        """
        Extension for recording nodes for removal during !merge
        process
        """
        return self.removed_node

    def replace(self, node):
        """
        Extension for recording nodes for replacement during !merge
        process
        """
        if isinstance(node, yaml.nodes.SequenceNode):
            node_list = self.construct_sequence(node)
            self.replaced_nodes.append(node_list)
            return node_list
        elif isinstance(node, yaml.nodes.MappingNode):
            mappings = self.construct_mapping(node)
            self.replaced_nodes.append(mappings)
            return mappings
        else:
            raise Exception(
                "Unsupported type for replace extension %s" % type(node))

    def join_text(self, node):
        """
        Extension for recording nodes for replacement during !merge
        process
        """
        text = self.construct_scalar(node)
        parts = [lpart.split("!}") for lpart in text.split("!{")]
        joined = []
        for part in parts:
            if len(part) == 1:
                joined.append(part[0])
            else:
                subloader = YamlStratusLoader(part[0], include_dirs=None,
                                              params=None)
                try:
                    obj = subloader.get_single_data()
                    joined.append(obj)
                finally:
                    subloader.dispose()

                joined.append(part[1])
        return {
            "Fn::Join": [
                "",
                joined
            ]
        }

    # Tag that merges two node to a single node
    # This one seems really useful
    def merge(self, node):
        """
        Extension for merging trees of nodes. Must have exactly two child nodes,
        one called 'startingFrom' and the other called 'mergeWith'.  The tree of
        nodes under 'startingFrom' is merged with the tree of nodes under
        'mergeWith'.

        Note that here we just record the nodes needing merge. The actual merge
        is done in a subsequent pass.
        """
        mappings = self.construct_mapping(node)

        if 'startingFrom' not in mappings:
            raise Exception(
                "!merge extension requires child node 'startingFrom'")
        if 'mergeWith' not in mappings:
            raise Exception("!merge extension requires child node 'mergeWith'")

        merge_node = MergeNode(mappings['startingFrom'], mappings['mergeWith'])

        self.merge_nodes.append(merge_node)

        return merge_node.target

    def param(self, node):
        """
        Extension for parameterizing
        """
        param_parts = self.construct_scalar(node).split(' ', 1)
        if not param_parts[0] in self.params_dict:
            if len(param_parts) > 1:
                return param_parts[1].strip('"')
            else:
                return ' '
        return self.params_dict[param_parts[0]]

    def apply_inheritance(self):
        """Iterate through all YAML nodes that need to be merged"""
        for merge_node in self.merge_nodes:
            self.merge_objects(merge_node)

    def merge_lists(self, src, override):
        """For merging to YAML nodes of type list"""
        merged = []

        if src is not None and override not in self.replaced_nodes:
            for val in src:
                merged.append(val)

        for val in override:
            merged.append(val)

        return merged

    def merge_dictionaries(self, src, override):
        """For merging to YAML nodes of type dictionary"""
        merged = dict()

        if override in self.replaced_nodes:
            # Ignore source
            src = None

        if src is not None:
            # Recursively merge the children of src with children of override
            for inner_key in src:
                if inner_key in override:
                    merged_children = self.merge_objects(
                        MergeNode(src[inner_key], override[inner_key]))
                else:
                    merged_children = src[inner_key]
                if merged_children is not None:
                    merged[inner_key] = merged_children

        for inner_key in override:
            # Add the children in override that are not in src
            if src is None or inner_key not in src:
                merged[inner_key] = override[inner_key]

        return merged

    def merge_objects(self, merge_node):
        """Recursively merge two YAML nodes"""
        if merge_node.is_merged:
            return merge_node.target
        merge_node.is_merged = True

        if self.removed_node == merge_node.merge_with:
            # Indicates a YAML node being removed
            return None

        if isinstance(merge_node.starting_from,
                      MergeNode) and not merge_node.starting_from.is_merged:
            # starting_from is a merge node that needs to be processed
            merge_node.starting_from = self.merge_objects(
                merge_node.starting_from)

        if isinstance(merge_node.merge_with,
                      MergeNode) and not merge_node.merge_with.is_merged:
            # merge_with is a merge node that needs to be processed
            merge_node.merge_with = self.merge_objects(merge_node.merge_with)

        # Check for merging of two YAML lists
        if isinstance(merge_node.starting_from, list) and isinstance(
                merge_node.merge_with, list):
            merge_node.target.extend(self.merge_lists(merge_node.starting_from,
                                                      merge_node.merge_with))
            return merge_node.target

        # Check for merging of two collections
        if hasattr(merge_node.starting_from, '__iter__') and hasattr(
                merge_node.merge_with, '__iter__'):
            if isinstance(merge_node.starting_from, list) or isinstance(
                    merge_node.merge_with, list):
                # A YAML list cannot be merged with a YAML dictionary
                raise Exception("Cannot merge lists with dictionaries")

            merged = self.merge_dictionaries(merge_node.starting_from,
                                             merge_node.merge_with)
            for key in merged:
                merge_node.target[key] = merged[key]
            return merge_node.target

        else:
            # Merging scalars
            if self.removed_node == merge_node.merge_with:
                return None
            return merge_node.merge_with


YamlStratusLoader.add_constructor('!include', YamlStratusLoader.include)

YamlStratusLoader.add_constructor('!include-base64',
                                  YamlStratusLoader.include_base64)

YamlStratusLoader.add_constructor('!merge', YamlStratusLoader.merge)

YamlStratusLoader.add_constructor('!remove', YamlStratusLoader.remove)

YamlStratusLoader.add_constructor('!replace', YamlStratusLoader.replace)

YamlStratusLoader.add_constructor('!param', YamlStratusLoader.param)

YamlStratusLoader.add_constructor('!jtext', YamlStratusLoader.join_text)


class YamlStratus(object):
    """
    Class for handling YAML stratus format

    """

    def __init__(self, include_dirs=None, params=None):
        """
        Parse a YAML stratus stream returning a dictionary

        :param include_dirs: list of search paths for includes (default [])
        :param params: dictionary of parameter name-values (default {})
        """
        self.include_dirs = include_dirs
        self.params = params

    def load(self, stream):
        """
        Parse a YAML stratus stream returning a dictionary

        :param stream: stream over the YAML Stratus document
        :rtype: dictionary
        """
        loader = YamlStratusLoader(stream, include_dirs=self.include_dirs,
                                   params=self.params)
        try:
            objs = loader.get_single_data()
            loader.apply_inheritance()
            return objs
        finally:
            loader.dispose()

    def load_all(self, stream):
        """
        Stream in multiple documents in YAML Stratus format returning
        an iteration over corresponding dictionaries

        :param stream: stream over the YAML Stratus document
        :rtype: iterator over dictionaries
        """
        loader = YamlStratusLoader(stream, include_dirs=self.include_dirs,
                                   params=self.params)
        try:
            while loader.check_data():
                objs = loader.get_data()
                loader.apply_inheritance()
                yield objs
        finally:
            loader.dispose()
